@@ -1,4 +1,4 @@
-/* dhcp-helper is Copyright (c) 2004 Simon Kelley
+/* dhcp-helper is Copyright (c) 2004,2006 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
 
 /* Author's email: simon@thekelleys.org.uk */
 
-#define VERSION "0.3"
+#define VERSION "0.7"
 
-#define COPYRIGHT "Copyright (C) 2004 Simon Kelley" 
+#define COPYRIGHT "Copyright (C) 2004-2006 Simon Kelley" 
 
 #include <sys/socket.h>
 #include <sys/ioctl.h>
@@ -32,15 +32,19 @@
 #include <errno.h>
 #include <pwd.h>
 #include <grp.h>
-#include <net/ethernet.h>
-#include <netpacket/packet.h>
-#include <netinet/ip.h>
 #include <netdb.h>
-#include <linux/sockios.h>
+#include <linux/types.h>
+#include <linux/capability.h>
+/* There doesn't seem to be a universally-available 
+   userpace header for this. */
+extern int capset(cap_user_header_t header, cap_user_data_t data);
+#include <sys/prctl.h>
+#include <net/if_arp.h>
 
 #define PIDFILE "/var/run/dhcp-helper.pid"
 #define USER "nobody"
 
+#define DHCP_CHADDR_MAX  16
 #define DHCP_SERVER_PORT 67
 #define DHCP_CLIENT_PORT 68
 #define BOOTREQUEST      1
@@ -58,36 +62,25 @@ struct interface {
   struct interface *next;
 };
 
-typedef unsigned char u8;
-typedef unsigned short u16;
-typedef unsigned int u32;
-
-struct udp_dhcp_packet {
-  struct ip ip;
-  struct udphdr {
-    u16 uh_sport;               /* source port */
-    u16 uh_dport;               /* destination port */
-    u16 uh_ulen;                /* udp length */
-    u16 uh_sum;                 /* udp checksum */
-  } udp;
+struct dhcp_packet_with_opts{
   struct dhcp_packet {
-    u8 op, htype, hlen, hops;
-    u32 xid;
-    u16 secs, flags;
+    unsigned char op, htype, hlen, hops;
+    unsigned int xid;
+    unsigned short secs, flags;
     struct in_addr ciaddr, yiaddr, siaddr, giaddr;
-    u8 chaddr[16], sname[64], file[128];
-    u8 options[312];
-  } data;
+    unsigned char chaddr[DHCP_CHADDR_MAX], sname[64], file[128];
+  } header;
+  unsigned char options[312];
 };
 
 
 int main(int argc, char **argv)
 {
-  int fd = -1, rawfd, opt, flags;
+  int fd = -1, opt;
   struct ifreq ifr;
   struct sockaddr_in saddr;
-  unsigned int buf_size = sizeof(struct udp_dhcp_packet);
-  struct udp_dhcp_packet *rawpacket;
+  size_t buf_size = sizeof(struct dhcp_packet_with_opts);
+  struct dhcp_packet *packet;
   struct namelist *interfaces = NULL, *except = NULL;
   struct interface *ifaces = NULL;
   struct namelist *servers = NULL;
@@ -206,7 +199,7 @@ int main(int argc, char **argv)
       exit(1); 
     }
 
-  if (!(rawpacket = malloc(buf_size)))
+  if (!(packet = malloc(buf_size)))
     {
       perror("dhcp-helper: cannot allocate buffer");
       exit(1);
@@ -219,9 +212,7 @@ int main(int argc, char **argv)
     }
   
   opt = 1;
-  if ((flags = fcntl(fd, F_GETFL, 0)) == -1 ||
-      fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1 ||
-      setsockopt(fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
+  if (setsockopt(fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
       setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(opt)) == -1)  
     {
       perror("dhcp-helper: cannot set options on DHCP socket");
@@ -237,18 +228,6 @@ int main(int argc, char **argv)
       exit(1);
     }
 
-  if ((rawfd = socket(PF_PACKET, SOCK_DGRAM, htons(ETHERTYPE_IP))) == -1)
-    {
-      perror("dhcp-helper: cannot create DHCP packet socket");
-      exit(1);
-    }
-
-  if (setsockopt(rawfd, SOL_SOCKET, SO_RCVBUF, &opt, sizeof(opt)) == -1)
-    {
-      perror("dhcp-helper: cannot set options on DHCP packet socket");
-      exit(1);
-    }
-
   if (!debug)
     {
       FILE *pidfile;
@@ -256,8 +235,24 @@ int main(int argc, char **argv)
       int i;
       gid_t dummy;
       struct group *gp;
-      
-      if (!ent_pw)
+      cap_user_header_t hdr = malloc(sizeof(*hdr));
+      cap_user_data_t data = malloc(sizeof(*data)); 
+
+      if (!hdr || !data)
+	{
+	  perror("dhcp-helper: cannot allocate memory");
+	  exit(1);
+	}
+
+      hdr->version = _LINUX_CAPABILITY_VERSION;
+      hdr->pid = 0; /* this process */
+      data->effective = data->permitted = data->inheritable =
+	(1 << CAP_NET_ADMIN) | (1 << CAP_SETGID) | (1 << CAP_SETUID);
+                  
+      /* Tell kernel to not clear capabilities when dropping root */
+      if (capset(hdr, data) == -1 || prctl(PR_SET_KEEPCAPS, 1) == -1)
+	ent_pw = NULL; /* must keep root if capset broken */
+      else if (!ent_pw)
 	{
 	  fprintf(stderr, "dhcp-helper: cannot find user %s\n", user);
 	  exit(1);
@@ -267,12 +262,12 @@ int main(int argc, char **argv)
          See Stevens section 12.4 */
 
       if (fork() != 0 )
-        exit(0);
+        _exit(0);
       
       setsid();
       
       if (fork() != 0)
-        exit(0);
+        _exit(0);
       
       chdir("/");
       umask(022); /* make pidfile 0644 */
@@ -287,64 +282,67 @@ int main(int argc, char **argv)
       umask(0);
 
       for (i=0; i<64; i++)        
-	if (i != rawfd && i != fd)
+	if (i != fd)
 	  close(i);
 
       setgroups(0, &dummy);
 
-      if ((gp = getgrgid(ent_pw->pw_gid)))
-	setgid(gp->gr_gid);
-      setuid(ent_pw->pw_uid);
+      if (ent_pw)
+	{
+	  if ((gp = getgrgid(ent_pw->pw_gid)))
+	    setgid(gp->gr_gid);
+	  setuid(ent_pw->pw_uid); 
+
+	  data->effective = data->permitted = 1 << CAP_NET_ADMIN;
+	  data->inheritable = 0;
+	  
+	  /* lose the setuid and setgid capbilities */
+	  capset(hdr, data);
+	}
     }
   
   while (1) {
-    fd_set rset;
     int iface_index;
     struct in_addr iface_addr;
-    unsigned int sz, size;    
-    struct dhcp_packet *header;
+    struct interface *iface;
+    ssize_t sz;
     struct msghdr msg;
-    struct iovec iov[1];
+    struct iovec iov;
     struct cmsghdr *cmptr;
+    struct in_pktinfo *pkt;
     union {
       struct cmsghdr align; /* this ensures alignment */
       char control[CMSG_SPACE(sizeof(struct in_pktinfo))];
     } control_u;
-    
-    FD_ZERO(&rset);
-    FD_SET(fd, &rset);
-
-    if (select(fd+1, &rset, NULL, NULL, NULL) == -1)
-      continue;
-
-    /* read size of waiting packet and expand buffer if necessary */
-    if (ioctl(fd, SIOCINQ, &size) != -1 &&
-	(size + sizeof(struct ip) + sizeof(struct udphdr)) > buf_size)
-      {
-	struct udp_dhcp_packet *newbuf = malloc(size + sizeof(struct ip) + sizeof(struct udphdr));
-	if (!newbuf)
-	  continue;
-	else
-	  {
-	    buf_size = size + sizeof(struct ip) + sizeof(struct udphdr);
-	    free(rawpacket);
-	    rawpacket = newbuf;
-	  }
-      }
-    
+        
     msg.msg_control = control_u.control;
     msg.msg_controllen = sizeof(control_u);
-    msg.msg_flags = 0;
-    msg.msg_name = NULL;
+    msg.msg_name = &saddr;
     msg.msg_namelen = 0;
-    msg.msg_iov = iov;
+    msg.msg_iov = &iov;
     msg.msg_iovlen = 1;
-    iov[0].iov_base = (char *)&rawpacket->data;
-    iov[0].iov_len = buf_size - (sizeof(struct ip) + sizeof(struct udphdr));
+    iov.iov_base = packet;
+    iov.iov_len = buf_size;
     
-    sz = recvmsg(fd, &msg, 0);
+    while (1) {
+      struct dhcp_packet *newbuf;
+      size_t newsz;
 
-    if (sz < (sizeof(*header) - sizeof(header->options)) || 
+      msg.msg_flags = 0;
+      while((sz = recvmsg(fd, &msg, MSG_PEEK)) == -1 && errno == EINTR);
+      
+      if (sz == -1 || !(msg.msg_flags & MSG_TRUNC) ||
+	  !(newbuf = realloc(packet, (newsz = buf_size + 100))))
+	break;
+      
+      iov.iov_base = packet = newbuf;
+      iov.iov_len = buf_size = newsz;
+    }
+    
+    while ((sz = recvmsg(fd, &msg, 0)) == -1 && errno == EINTR);
+    
+    if ((msg.msg_flags & MSG_TRUNC) ||
+	sz < (ssize_t)(sizeof(struct dhcp_packet)) || 
 	msg.msg_controllen < sizeof(struct cmsghdr))
       continue;
     
@@ -355,25 +353,25 @@ int main(int argc, char **argv)
   
     if (!(ifr.ifr_ifindex = iface_index) || ioctl(fd, SIOCGIFNAME, &ifr) == -1)
       continue;
-    	
+    	 
     ifr.ifr_addr.sa_family = AF_INET;
     if (ioctl(fd, SIOCGIFADDR, &ifr) == -1)
       continue;
     else
       iface_addr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;
     
-    header = (struct dhcp_packet *)&rawpacket->data;
-    
     /* last ditch loop squashing. */
-    if ((header->hops++) > 20)
+    if ((packet->hops++) > 20)
       continue;
 
-    if (header->op == BOOTREQUEST)
+    if (packet->hlen > DHCP_CHADDR_MAX)
+      continue;
+
+    if (packet->op == BOOTREQUEST)
       {
 	/* message from client */
 	struct namelist *tmp;
-	struct interface *iface;
-
+	
 	/* packets from networks we are broadcasting _too_
 	   are explicitly not allowed to be forwarded _from_ */
 	for (tmp = servers; tmp; tmp = tmp->next)
@@ -399,11 +397,11 @@ int main(int argc, char **argv)
 	  }
 	
 	/* already gatewayed ? */
-	if (header->giaddr.s_addr)
+	if (packet->giaddr.s_addr)
 	  {
 	    /* if so check if by us, to stomp on loops. */
 	    for (iface = ifaces; iface; iface = iface->next)
-	      if (iface->addr.s_addr == header->giaddr.s_addr)
+	      if (iface->addr.s_addr == packet->giaddr.s_addr)
 		break;
 	    if (iface)
 	      continue;
@@ -411,7 +409,7 @@ int main(int argc, char **argv)
 	else
 	  {
 	    /* plug in our address */
-	    header->giaddr = iface_addr;
+	    packet->giaddr = iface_addr;
 	  }
 
 	/* send to all configured servers. */
@@ -429,7 +427,8 @@ int main(int argc, char **argv)
 	      saddr.sin_addr = tmp->addr;
 	    
 	    saddr.sin_port = htons(DHCP_SERVER_PORT);
-	    sendto(fd, &rawpacket->data, sz, 0, (struct sockaddr *)&saddr, sizeof(saddr));
+	    while(sendto(fd, packet, sz, 0, (struct sockaddr *)&saddr, sizeof(saddr)) == -1 &&
+		  errno == EINTR);
 	  }
 
 	/* build address->interface index table for returning answers */
@@ -449,89 +448,60 @@ int main(int argc, char **argv)
 	    iface->index = iface_index;
 	  }
       }
-    else if (header->op == BOOTREPLY)
+    else if (packet->op == BOOTREPLY)
       { 
-	/* packet from server send back to client */
-	if (header->ciaddr.s_addr)
+	/* packet from server send back to client */	
+	saddr.sin_port = htons(DHCP_CLIENT_PORT);
+	msg.msg_control = NULL;
+	msg.msg_controllen = 0;
+	msg.msg_namelen = sizeof(saddr);
+	iov.iov_len = sz;
+			   
+	/* look up interface index in cache */
+	for (iface = ifaces; iface; iface = iface->next)
+	  if (iface->addr.s_addr == packet->giaddr.s_addr)
+	    break;
+	
+	if (!iface)
+	  continue;
+            
+	if (packet->ciaddr.s_addr)
+	  saddr.sin_addr = packet->ciaddr;
+	else if (ntohs(packet->flags) & 0x8000 || packet->hlen > 14)
 	  {
-	    /* client has configured address, this is easy (might never happen) */
-	    saddr.sin_addr = header->ciaddr;
-	    saddr.sin_port = htons(DHCP_CLIENT_PORT);
-	    sendto(fd, &rawpacket->data, sz, 0, (struct sockaddr *)&saddr, sizeof(saddr));
+	    /* broadcast to 255.255.255.255 */
+	    msg.msg_controllen = sizeof(control_u);
+	    msg.msg_control = control_u.control;
+	    cmptr = CMSG_FIRSTHDR(&msg);
+	    saddr.sin_addr.s_addr = INADDR_BROADCAST;
+	    pkt = (struct in_pktinfo *)CMSG_DATA(cmptr);
+	    pkt->ipi_ifindex = iface->index;
+	    pkt->ipi_spec_dst.s_addr = 0;
+	    msg.msg_controllen = cmptr->cmsg_len = CMSG_LEN(sizeof(struct in_pktinfo));
+	    cmptr->cmsg_level = SOL_IP;
+	    cmptr->cmsg_type = IP_PKTINFO;
 	  }
-	else
+	else 
 	  {
-	    /* client not configured and cannot reply to ARP. Have to send
-	       directly to its MAC address or broadcast. */
-
-	    struct sockaddr_ll dest;
-	    struct interface *iface;
-	    u32 i, sum;
-	    
-	    /* look up interface index in cache */
-	    for (iface = ifaces; iface; iface = iface->next)
-	      if (iface->addr.s_addr == header->giaddr.s_addr)
-		break;
-			    
-	    if ((header->hlen > 8) || !iface)
-	      continue;
-	    
-	    dest.sll_ifindex =  iface->index;
-	    dest.sll_family = AF_PACKET;
-	    dest.sll_halen =  header->hlen;
-	    dest.sll_protocol = htons(ETHERTYPE_IP);
-	     
-	    if (ntohs(header->flags) & 0x8000)
+	    /* client not configured and cannot reply to ARP. 
+	       Insert arp entry direct.*/
+	    saddr.sin_addr = packet->yiaddr;
+	    ifr.ifr_ifindex = iface->index;
+	    if (ioctl(fd, SIOCGIFNAME, &ifr) != -1)
 	      {
-		memset(dest.sll_addr, 255,  header->hlen);
-		rawpacket->ip.ip_dst.s_addr = INADDR_BROADCAST;
+		struct arpreq req;
+		*((struct sockaddr_in *)&req.arp_pa) = saddr;
+		req.arp_ha.sa_family = packet->htype;
+		memcpy(req.arp_ha.sa_data, packet->chaddr, packet->hlen);
+		strncpy(req.arp_dev, ifr.ifr_name, 16);
+		req.arp_flags = ATF_COM;
+		ioctl(fd, SIOCSARP, &req);
 	      }
-	    else
-	      {
-		memcpy(dest.sll_addr, header->chaddr, header->hlen); 
-		rawpacket->ip.ip_dst.s_addr = header->yiaddr.s_addr;
-	      }
-	    
-	    rawpacket->ip.ip_p = IPPROTO_UDP;
-	    rawpacket->ip.ip_src.s_addr = iface_addr.s_addr;
-	    rawpacket->ip.ip_len = htons(sizeof(struct ip) + 
-					 sizeof(struct udphdr) +
-					 sz) ;
-	    rawpacket->ip.ip_hl = sizeof(struct ip) / 4;
-	    rawpacket->ip.ip_v = IPVERSION;
-	    rawpacket->ip.ip_tos = 0;
-	    rawpacket->ip.ip_id = htons(0);
-	    rawpacket->ip.ip_off = htons(0x4000); /* don't fragment */
-	    rawpacket->ip.ip_ttl = IPDEFTTL;
-	    rawpacket->ip.ip_sum = 0;
-	    
-	    for (sum = 0, i = 0; i < sizeof(struct ip) / 2; i++)
-	      sum += ((u16 *)&rawpacket->ip)[i];
-	    while (sum>>16)
-	      sum = (sum & 0xffff) + (sum >> 16);  
-	    rawpacket->ip.ip_sum = (sum == 0xffff) ? sum : ~sum;
-	    
-	    rawpacket->udp.uh_sport = htons(DHCP_SERVER_PORT);
-	    rawpacket->udp.uh_dport = htons(DHCP_CLIENT_PORT);
-	    ((u8 *)&rawpacket->data)[sz] = 0; /* for checksum, in case length is odd. */
-	    
-	    rawpacket->udp.uh_sum = 0;
-	    rawpacket->udp.uh_ulen = sum = htons(sizeof(struct udphdr) + sz);
-	    sum += htons(IPPROTO_UDP);
-	    for (i = 0; i < 4; i++)
-	      sum += ((u16 *)&rawpacket->ip.ip_src)[i];
-	    for (i = 0; i < (sizeof(struct udphdr) + sz + 1) / 2; i++)
-	      sum += ((u16 *)&rawpacket->udp)[i];
-
-	    while (sum>>16)
-	      sum = (sum & 0xffff) + (sum >> 16);
-	    rawpacket->udp.uh_sum = (sum == 0xffff) ? sum : ~sum;
-
-	    sendto(rawfd, rawpacket, ntohs(rawpacket->ip.ip_len), 
-		   0, (struct sockaddr *)&dest, sizeof(dest));
 	  }
+
+	while (sendmsg(fd, &msg, 0) == -1 && errno == EINTR);
       }
   }
-
 }
+    
 	    
